"""Test the validators module."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pytest
from jsonschema.exceptions import ValidationError

from diameter_synthesis import validators


def test_defaults(config, model_params):
    """Test the validation of JSON schemas."""
    mtype = "L5_TPC:A"
    model = "generic"

    validators.validate_model_params(model_params[model][mtype])
    validators.validate_config(config[model])

    config[model]["features"] = {"FEATURE 1": {"apical": 0.2}}
    validators.validate_config(config[model])

    config[model]["features"] = {"FEATURE 2": {"UNKNOWN": 0.2}}
    with pytest.raises(ValidationError):
        validators.validate_config(config[model])



"""Prepare the tests."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=redefined-outer-name
import json
import shutil
from pathlib import Path

import morphio.mut
import neurom as nm
import numpy as np
import pandas as pd
import pytest
from morphio import PointLevel
from morphio import SectionType


@pytest.fixture(autouse=True)
def set_random():
    """Path to the test directory."""
    np.random.seed(0)


@pytest.fixture
def test_path():
    """Path to the test directory."""
    return Path(__file__).absolute().parent


@pytest.fixture
def test_data_path(test_path):
    """Path to the test data directory."""
    return test_path / "data"


@pytest.fixture
def expected_images_path(test_path):
    """Path to the expected images directory."""
    return test_path / "expected_images"


@pytest.fixture
def config_path(test_data_path):
    """Path to the config file."""
    return test_data_path / "config.json"


@pytest.fixture
def config(config_path):
    """The config used in tests."""
    with open(config_path, encoding="utf-8") as filename:
        return json.load(filename)


@pytest.fixture
def model_params_path(test_data_path):
    """The path to the model parameters."""
    return test_data_path / "model_params.json"


@pytest.fixture
def model_params(model_params_path):
    """The model parameters."""
    with open(model_params_path, encoding="utf-8") as filename:
        return json.load(filename)


@pytest.fixture
def model_data(test_data_path):
    """The model data."""
    with open(test_data_path / "model_data.json", "r", encoding="utf-8") as filename:
        return json.load(filename)


@pytest.fixture
def neuron_path(test_data_path):
    """The path to the tested morphology."""
    return test_data_path / "C030796A-P3_lite.h5"


@pytest.fixture
def neuron(neuron_path):
    """The tested morphology."""
    return morphio.mut.Morphology(neuron_path)


@pytest.fixture
def neuron_diametrized_path(test_data_path):
    """The path to the diametrized morphology."""
    return test_data_path / "C030796A-P3_lite_diametrized.h5"


@pytest.fixture
def neuron_diametrized(neuron_diametrized_path):
    """The diametrized morphology."""
    return morphio.mut.Morphology(neuron_diametrized_path)


@pytest.fixture
def single_pop(neuron_path):
    """A population with one morphology."""
    return [nm.load_morphology(neuron_path)]


@pytest.fixture
def single_pop_diametrized(neuron_diametrized_path):
    """A population with one diametrized morphology."""
    return [nm.load_morphology(neuron_diametrized_path)]


@pytest.fixture
def single_neurite(neuron_path):
    """A neurite."""
    neuron = nm.load_morphology(neuron_path)
    yield neuron.neurites[0]


@pytest.fixture
def single_pop_neurondb_dat_path(test_data_path):
    """The path to the neurondb.dat file."""
    return test_data_path / "neurondb.dat"


@pytest.fixture
def single_pop_neurondb(single_pop_neurondb_dat_path):
    """The DF of the singme population."""
    df = pd.read_csv(single_pop_neurondb_dat_path, header=None, sep=r"\s+")
    df.rename(columns={0: "morphology", 1: "layer", 2: "mtype"}, inplace=True)
    return df


@pytest.fixture
def single_pop_data_dir(tmpdir, single_pop_neurondb, neuron_path):
    """Prepare test data."""
    single_pop_dir = tmpdir / "single_pop"

    single_pop_dir.mkdir()
    single_pop_neurondb.to_csv(single_pop_dir / "neurondb.dat", sep=" ", header=False, index=False)
    shutil.copyfile(neuron_path, single_pop_dir / neuron_path.name)

    return single_pop_dir


@pytest.fixture
def single_pop_diametrized_data_dir(
    tmpdir, single_pop_neurondb, neuron_path, neuron_diametrized_path
):
    """Prepare test data with diametrized morphologies."""
    single_pop_diametrized_dir = tmpdir / "single_pop_diametrized"
    single_pop_diametrized_dir.mkdir()
    single_pop_neurondb.loc[0, "morphology"] += "_diametrized"
    single_pop_neurondb.to_csv(
        single_pop_diametrized_dir / "neurondb.dat", sep=" ", header=False, index=False
    )
    shutil.copyfile(neuron_diametrized_path, single_pop_diametrized_dir / neuron_path.name)

    return single_pop_diametrized_dir


@pytest.fixture
def empty_build_result():
    """The result of an empty build."""
    return (
        {
            "diameter_power_relation": {"apical": {}, "basal": {}},
            "sibling_ratios": {"apical": {}, "basal": {}},
            "tapers": {"apical": {}, "basal": {}},
            "terminal_diameters": {"apical": {}, "basal": {}},
            "trunk_diameters": {"apical": {}, "basal": {}},
        },
        {
            "diameter_power_relation": {"apical": [], "basal": []},
            "sibling_ratios": {"apical": [], "basal": []},
            "tapers": {"apical": [], "basal": []},
            "terminal_diameters": {"apical": [], "basal": []},
            "trunk_diameters": {"apical": [], "basal": []},
        },
    )


@pytest.fixture
def small_morph():
    """A small morphology used in tests."""
    N = 3
    dx = 100
    dy = 100
    Ndy = (N - 1) * dy

    nrn = morphio.mut.Morphology()

    # Define a basal with only one section
    x0_b1 = 0
    nrn.append_root_section(
        PointLevel([[x0_b1, i * dy, 0] for i in range(N)], [0.2] * N), SectionType(3)
    )

    # Define a basal with several sections
    x0_b2 = 0.1
    root = nrn.append_root_section(
        PointLevel([[x0_b2, i * dy, 0] for i in range(N)], [0.2] * N), SectionType(3)
    )
    root.append_section(
        PointLevel([[x0_b2, Ndy + i * dy, 0] for i in range(N)], [0.2] * N), SectionType(3)
    )
    root.append_section(
        PointLevel([[x0_b2 + i * -0.1, Ndy + i * dy, 0] for i in range(N)], [0.2] * N),
        SectionType(3),
    )

    # Define the apical with several bifurcations
    x0_ap = 0.2
    apical_root = nrn.append_root_section(
        PointLevel([[x0_ap, i * dy, 0] for i in range(N)], [0.2] * N), SectionType(4)
    )
    apical_root.append_section(
        PointLevel([[x0_ap + i * dx, Ndy + i * dy, 0] for i in range(N)], [0.2] * N), SectionType(4)
    )
    major = apical_root.append_section(
        PointLevel([[x0_ap, Ndy + i * dy, 0] for i in range(N)], [0.2] * N), SectionType(4)
    )

    major.append_section(
        PointLevel([[x0_ap + i * dx, 2 * Ndy + i * dy, 0] for i in range(N)], [0.2] * N),
        SectionType(4),
    )
    major = major.append_section(
        PointLevel([[x0_ap, 2 * Ndy + i * dy, 0] for i in range(N)], [0.2] * N), SectionType(4)
    )

    major.append_section(
        PointLevel([[x0_ap + i * dx, 3 * Ndy + i * dy, 0] for i in range(N)], [0.2] * N),
        SectionType(4),
    )
    major = major.append_section(
        PointLevel([[x0_ap, 3 * Ndy + i * dy, 0] for i in range(N)], [0.2] * N), SectionType(4)
    )

    major.append_section(
        PointLevel([[x0_ap + i * dx, 4 * Ndy + i * dy, 0] for i in range(N)], [0.2] * N),
        SectionType(4),
    )
    major = major.append_section(
        PointLevel([[x0_ap, 4 * Ndy + i * dy, 0] for i in range(N)], [0.2] * N), SectionType(4)
    )

    # Define an axon with one section
    x0_ax1 = 0
    y0_ax1 = -0.1
    nrn.append_root_section(
        PointLevel([[x0_ax1, y0_ax1 - i * dy, 0] for i in range(N)], [0.2] * N), SectionType(2)
    )

    # Define an axon with one bifurcation
    x0_ax2 = 0.1
    y0_ax2 = -0.1
    root = nrn.append_root_section(
        PointLevel([[x0_ax2, y0_ax2 - i * dy, 0] for i in range(N)], [0.2] * N), SectionType(2)
    )
    root.append_section(
        PointLevel([[x0_ax2, y0_ax2 - Ndy - i * dy, 0] for i in range(N)], [0.2] * N),
        SectionType(3),
    )
    root.append_section(
        PointLevel([[x0_ax2 + i * -0.1, y0_ax2 - Ndy - i * dy, 0] for i in range(N)], [0.2] * N),
        SectionType(3),
    )

    # Define soma according to the root sections
    nrn.soma.points = [
        [x0_b1, 0, 0],
        [x0_b2, 0, 0],
        [x0_ap, 0, 0],
        [x0_ax2, y0_ax2, 0],
        [x0_ax1, y0_ax1, 0],
    ]
    nrn.soma.diameters = [0] * len(nrn.soma.points)

    return nrn



"""Morphology extraction functions."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from functools import lru_cache

import neurom as nm
import numpy as np
from neurom.core.morphology import Section
from neurom.core.morphology import iter_sections

from diameter_synthesis.exception import DiameterSynthesisError
from diameter_synthesis.utils import _get_diameters
from diameter_synthesis.utils import _get_mean_diameter

L = logging.getLogger(__name__)


def _segment_lengths(section):
    """Lengths of segments of a section (morphio only)."""
    points = section.points
    vectors = np.diff(points, axis=0)
    return np.linalg.norm(vectors, axis=1)


@lru_cache(maxsize=None)
def sec_length(section):
    """Length of a section (morphio only)."""
    return _segment_lengths(section).sum()


@lru_cache(maxsize=None)
def lengths_from_origin(section):
    """Path lengths from first point of section (morphio only)."""
    return np.insert(np.cumsum(_segment_lengths(section)), 0, 0)


@lru_cache(maxsize=None)
def partition_asymmetry_length(section):
    """Compute partition asymmetry with lengths (morphio).

    WARNING: This function is not complete, and requires to divide by total length of dendrite,
    which is done later in the code as we don't have access to whole dendrite here.
    """
    asymmetry_length = sum([sec_length(section) for section in section.children[0].iter()])
    try:
        asymmetry_length -= sum([sec_length(section) for section in section.children[1].iter()])
    except IndexError as exc:
        raise DiameterSynthesisError(
            "Bifurcation point with single child, consider using neuror sanitize"
        ) from exc
    return abs(asymmetry_length)


@lru_cache(maxsize=None)
def n_children_downstream(section):
    """Get the number of children of a section (morphio only)."""
    return sum(1 for _ in section.iter())


def get_total_length(neurite):
    """Get total length or a neurite (morphio only)."""
    return sum(map(sec_length, neurite))


def compute_sibling_ratios(neurite, method="mean", attribute_name=None, bounds=None):
    """Compute the sibling ratios of a neurite (neurom only)."""
    sibling_ratios_values = nm.get("sibling_ratios", neurite, method=method)

    if not bounds:
        bounds = [0, 1 + 1e-5]
    return add_additional_attributes(
        sibling_ratios_values, neurite, attribute_name=attribute_name, bounds=bounds
    )


def compute_diameter_power_relation(neurite, method="mean", attribute_name=None, bounds=None):
    """Compute the Rall deviation the diameter of the segments of a tree (neurom only)."""
    diameter_power_relation_values = nm.get("diameter_power_relations", neurite, method=method)
    if not bounds:
        bounds = [0.0, 100.0]

    return add_additional_attributes(
        diameter_power_relation_values,
        neurite,
        attribute_name=attribute_name,
        bounds=bounds,
    )


def diameter_power_relation_factor(diameter_power_relation, sibling_ratio):
    """Compute the reduction factor for bifurcation diameter."""
    if sibling_ratio > 0:
        return ((1.0 + sibling_ratio ** -1.5) / diameter_power_relation) ** (2.0 / 3.0)
    return 1.0


def terminal_diameters(neurite, method="mean", threshold=1.0, attribute_name=None, bounds=None):
    """Compute the terminal diameters of a neurite (neurom only)."""
    diameters = []
    for section in iter_sections(neurite):
        diameters += list(_get_diameters(section))
    mean_diameter = np.mean(diameters)

    if method == "mean":
        term_diam = [
            np.mean(_get_diameters(t))
            for t in Section.ileaf(next(iter_sections(neurite)))
            if np.mean(_get_diameters(t)) < threshold * mean_diameter
        ]

    elif method == "first":
        term_diam = [
            _get_diameters(t)[-1]
            for t in Section.ileaf(next(iter_sections(neurite)))
            if _get_diameters(t)[-1] < threshold * mean_diameter
        ]

    else:
        raise DiameterSynthesisError("Method for singling computation not understood!")

    if not bounds:
        bounds = [0, 100]
    return add_additional_attributes(
        term_diam, neurite, attribute_name=attribute_name, bounds=bounds
    )


def min_diameter(neurite, attribute_name=None, bounds=None):
    """Get the min diameter of a neurite (neurom only)."""
    min_diam = min(map(_get_mean_diameter, iter_sections(neurite)))

    if not bounds:
        bounds = [0, 100]
    return add_additional_attributes(
        [min_diam], neurite, attribute_name=attribute_name, bounds=bounds
    )


def max_diameter(neurite, attribute_name=None, bounds=None):
    """Get the max diameter of a neurite (neurom only)."""
    max_diam = max(map(_get_mean_diameter, iter_sections(neurite)))

    if not bounds:
        bounds = [0, 100]
    return add_additional_attributes(
        [max_diam], neurite, attribute_name=attribute_name, bounds=bounds
    )


def trunk_diameter(neurite, attribute_name=None, bounds=None, method="last"):
    """Get the trunc diameters (neurom only)."""
    if method == "mean":
        trunk_diam = _get_mean_diameter(neurite.root_node)
    if method == "first":
        trunk_diam = _get_diameters(neurite.root_node)[0]
    if method == "last":
        trunk_diam = _get_diameters(neurite.root_node)[-1]

    if not bounds:
        bounds = [0, 100]
    return add_additional_attributes(
        [trunk_diam], neurite, attribute_name=attribute_name, bounds=bounds
    )


def taper(neurite, params, attribute_name=None):
    """Get the taper rates (neurom only)."""
    if params is None:
        raise DiameterSynthesisError("Please provide parameters for taper rate computations")
    tapers = nm.get("section_taper_rates", neurite)
    return add_additional_attributes(
        np.array(tapers),
        neurite,
        attribute_name=attribute_name,
        bounds=[params["min"], params["max"]],
    )


def get_additional_attribute(
    attribute_name, neurite=None, section=None
):  # noqa, pylint: disable=too-many-return-statements,too-many-branches
    """Return the value of an additional attribute of a parameter, given a neurite or a section."""
    section_only = section is not None and neurite is None
    neurite_only = neurite is not None and section is None
    if attribute_name in ["asymmetry", "asymmetry_threshold"]:
        if neurite_only:
            out = nm.get("partition_asymmetry", neurite, variant="length")
            return np.array(out)
        if section_only:
            return partition_asymmetry_length(section)

    if attribute_name == "asymmetry_pair":
        if section_only:
            return [nm.features.bifurcation.partition_pair(section)]
        raise DiameterSynthesisError("Please provide only a section")

    if attribute_name == "tot_length":
        if neurite_only:
            return [nm.get("total_length", neurite)]
        raise DiameterSynthesisError("Please provide only a neurite")

    if attribute_name == "max_path":
        if neurite_only:
            return [np.max(nm.get("section_path_distances", neurite))]
        raise DiameterSynthesisError("Please provide only a neurite")

    if attribute_name == "max_branch":
        if neurite_only:
            return [np.max(nm.get("section_branch_orders", neurite))]
        raise DiameterSynthesisError("Please provide only a neurite")

    if attribute_name == "root_strahler":
        if neurite_only:
            return [nm.get("section_strahler_orders", neurite)[0]]
        raise DiameterSynthesisError("Please provide only a neurite")

    if attribute_name == "sibling":
        if neurite_only:
            return compute_sibling_ratios(neurite, method="mean")
        if section_only:
            return nm.features.bifurcation.sibling_ratio(section, method="mean")

    raise DiameterSynthesisError(
        "Please provide a valid attribute_name and provide either a neurite or a section, not both"
    )


def add_additional_attributes(data, neurite, attribute_name=None, bounds=None):
    """From a data and a type of additional, return both within some bounds."""
    if not bounds:
        bounds = [-1000, 1000]

    if attribute_name is None:
        data = np.array(data)
        return list(data[(data >= bounds[0]) & (data < bounds[1])])

    return [
        [d, s]
        for d, s in zip(data, get_additional_attribute(attribute_name, neurite=neurite))
        if bounds[0] <= d < bounds[1]
    ]



"""Click app for the diameter-synthesis package."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=import-outside-toplevel,redefined-outer-name
import logging
import os
from pathlib import Path

import click
import morphio
from tqdm import tqdm

from .utils import create_morphologies_dict

morphio.set_maximum_warnings(0)
L = logging.getLogger(__name__)
logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))


@click.group()
def cli():
    """Cli to learn and generate diameters of neurons."""


@cli.command("run_models")
@click.argument("config_file", type=click.Path(exists=True))
@click.option("--plot", is_flag=True)
@click.option("--ext", default=".png")
def run_models(config_file, plot=False, ext=".png"):
    """Run the model extraction from config file."""
    from .main import run_models

    run_models(config_file, plot=plot, ext=ext)


@cli.command("run_diameters")
@click.argument("config_file", type=click.Path(exists=True))
@click.argument("models_params_file", type=click.Path(exists=True))
def run_diameters(config_file, models_params_file):
    """Build new diameters from config file and diameter model."""
    from .main import run_diameters

    run_diameters(config_file, models_params_file)


@cli.command("plot_diff")
@click.option("--orig-path", help="Path to original cells", required=True)
@click.option("--diam-path", help="Path to diametrized cells", required=True)
@click.option("-o", "--out-dir", help="Directory to output the analysis results", required=True)
@click.option("-n", "--ncells", default=10, help="Max number of cells to plot")
@click.option("-e", "--ext", default=".png", help="Figures extension")
def plot_diff(orig_path, diam_path, out_dir, ncells=None, ext=".png"):
    """Plot original and new neurons as well as their differences."""
    from .plotting import plot_diameter_diff

    neurite_types = ["basal", "apical", "axon"]
    import neurom as nm

    morphologies_dict = create_morphologies_dict(orig_path)
    out_dir = Path(out_dir) / "diffs"
    if not out_dir.exists():
        out_dir.mkdir(parents=True, exist_ok=True)

    for mtype in morphologies_dict:
        if len(morphologies_dict) > 1:
            L.info("Plot mtype %s", mtype)
            plot_folder_mtype = out_dir / mtype
            if not plot_folder_mtype.exists:
                os.mkdir(plot_folder_mtype)
        else:
            plot_folder_mtype = out_dir

        if ncells is not None:
            neurons = morphologies_dict[mtype][: int(ncells)]
        else:
            neurons = morphologies_dict[mtype]

        for neuron in tqdm(neurons):
            neuron_new = nm.load_morphology(Path(diam_path) / neuron.name)
            plot_diameter_diff(neuron, neuron_new, neurite_types, plot_folder_mtype, ext=ext)


@cli.command("run_analysis")
@click.option("--orig-path", help="Path to original cells", required=True)
@click.option("--diam-path", help="Path to diametrized cells", required=True)
@click.option("--mtypes-file", help="Path to mtypes file", required=False)
@click.option("-o", "--out-dir", help="Directory to output the analysis results", required=True)
@click.option("--cumulative", help="Cumulative distribution plots", is_flag=True)
@click.option("--individual", help="Output a plot for each neuron", is_flag=True)
@click.option("--violin", help="Violin distribution plots", is_flag=True)
@click.option("-e", "--ext", default=".png", help="Figures extension")
def run_analysis(
    orig_path, diam_path, out_dir, cumulative, individual, violin, mtypes_file=None, ext=".png"
):
    """Produce figures for validation/analysis."""
    if not cumulative and not violin:
        raise ValueError("Should at least set one of '--cumulative' or '--violin' to True.")
    if cumulative:
        from .plotting import cumulative_analysis

        for neurite_type in ["basal", "axon", "apical"]:
            cumulative_analysis(
                orig_path,
                diam_path,
                Path(out_dir) / neurite_type,
                individual,
                mtypes_file=mtypes_file,
                neurite_types=[neurite_type],
                ext=ext,
            )
    if violin:
        from .plotting import violin_analysis

        violin_analysis(orig_path, diam_path, out_dir, mtypes_file=mtypes_file)



"""Some tools used in tests."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json

import neurom as nm
import numpy as np
from numpy.testing import assert_allclose


def _copy_diameters(neuron_a, neuron_b):
    for section_a, section_b in zip(nm.iter_sections(neuron_a), nm.iter_sections(neuron_b)):
        section_a.diameters = section_b.diameters


def _compare_diameters(neuron_a, neuron_b, **kwargs):
    """Test if all diameters are the same."""
    errs = []
    for section_a, section_b in zip(neuron_a.sections, neuron_b.sections):
        try:
            assert_allclose(
                neuron_a.sections[section_a].diameters,
                neuron_b.sections[section_b].diameters,
                err_msg=f"Error for section {section_a}",
                **kwargs,
            )
        except AssertionError as e:
            errs.append(e)

    if errs:
        raise AssertionError("\n".join([i.args[0] for i in errs]))


def nested_round(obj, precision=6):
    """Round all floats (recursively) in a nested dictionary."""
    # pylint: disable=too-many-return-statements
    if precision is None:
        precision = 0
    if isinstance(obj, np.ndarray):
        return [nested_round(i, precision) for i in obj.tolist()]
    if isinstance(obj, np.floating):
        return round(float(obj), precision)
    if isinstance(obj, np.integer):
        return round(int(obj), precision)
    if isinstance(obj, float):
        return round(obj, precision)
    if isinstance(obj, dict):
        return dict((k, nested_round(v, precision)) for k, v in obj.items())
    if isinstance(obj, (list, tuple)):
        return [nested_round(i, precision) for i in obj]
    return obj


def compare_dicts(ref, test, precision=None):
    """Compare two dictionaries."""
    if precision is not None:
        ref = nested_round(ref, precision)
        test = nested_round(test, precision)
    return json.dumps(ref, sort_keys=True) == json.dumps(test, sort_keys=True)



"""Test the cli module."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
from pathlib import Path

import pytest
from click.testing import CliRunner

from diameter_synthesis import cli

from .testing_tools import compare_dicts


def test_run_models(tmpdir, single_pop_data_dir, single_pop_diametrized_data_dir, config):
    """Test the run_models entry point."""
    # Prepare inputs
    extract_models_params = config["generic"]
    extract_models_params["mtypes_file"] = str(single_pop_data_dir / "neurondb.dat")
    extract_models_params["morph_path"] = str(single_pop_data_dir)
    extract_models_params["new_morph_path"] = str(single_pop_diametrized_data_dir)
    extract_models_params["models_params_file"] = str(tmpdir / "model_params_mtypes.json")
    extract_models_params["fig_folder"] = str(tmpdir / "model_figures")
    extract_models_params["n_cpu"] = 1
    extract_models_params["neurite_types"] = ["basal", "apical"]

    config_file = str(tmpdir / "diametrizer_params.json")
    with open(config_file, "w", encoding="utf-8") as json_file:
        json.dump(extract_models_params, json_file, sort_keys=True, indent=4)

    # Run with CLI
    runner = CliRunner()
    runner.invoke(cli.cli, ["run_models", config_file], catch_exceptions=False)

    # Check results
    with open(extract_models_params["models_params_file"], "r", encoding="utf-8") as json_file:
        res = json.load(json_file)

    assert list(res.keys()) == ["generic"]
    assert list(res["generic"].keys()) == ["L5_TPC:A"]

    # Check only diameter_power_relation entry
    compare_dicts(
        res["generic"]["L5_TPC:A"]["diameter_power_relation"],
        {
            "apical": {
                "distribution": "exponnorm",
                "params": {
                    "a": 0.3,
                    "loc": 10.324635885438347,
                    "max": 21.302095794677733,
                    "min": 3.745589017868042,
                    "num_value": 4,
                    "scale": 7.39024091228236,
                },
                "sequential": "asymmetry_threshold",
            },
            "basal": {
                "distribution": "exponnorm",
                "params": {
                    "a": 4.0,
                    "loc": 2.081371164504187,
                    "max": 13.171166133880611,
                    "min": 2.4313707590103153,
                    "num_value": 10,
                    "scale": 0.8153109268097628,
                },
                "sequential": "asymmetry_threshold",
            },
        },
        precision=6,
    )


def test_run_diameters(tmpdir, single_pop_data_dir, config, model_params_path):
    """Test the run_diameters entry point."""
    # Prepare inputs
    res_path = Path(tmpdir / "new_morphologies")
    extract_models_params = config["generic"]
    extract_models_params["mtypes_file"] = str(single_pop_data_dir / "neurondb.dat")
    extract_models_params["morph_path"] = str(single_pop_data_dir)
    extract_models_params["n_cpu"] = 1
    extract_models_params["neurite_types"] = ["basal", "apical"]
    extract_models_params["new_morph_path"] = str(res_path)

    config_file = str(tmpdir / "diametrizer_params.json")
    with open(config_file, "w", encoding="utf-8") as json_file:
        json.dump(extract_models_params, json_file, sort_keys=True, indent=4)

    # Run with CLI
    runner = CliRunner()
    runner.invoke(
        cli.cli, ["run_diameters", config_file, str(model_params_path)], catch_exceptions=False
    )

    # Check results
    assert [i.name for i in res_path.iterdir()] == ["C030796A-P3_lite.h5"]


def test_plot_diff(tmpdir, single_pop_data_dir, single_pop_diametrized_data_dir):
    """Test the plot_diff entry point."""
    res_path = Path(tmpdir / "figures")

    # Run with CLI
    runner = CliRunner()
    runner.invoke(
        cli.cli,
        [
            "plot_diff",
            "--orig-path",
            str(single_pop_data_dir),
            "--diam-path",
            str(single_pop_diametrized_data_dir),
            "--out-dir",
            str(res_path),
            "--ext",
            ".pdf",
        ],
        catch_exceptions=False,
    )

    # Check results
    assert [i.name for i in (res_path / "diffs").iterdir()] == ["C030796A-P3_lite.pdf"]


def test_run_analysis(tmpdir, single_pop_data_dir, single_pop_diametrized_data_dir):
    """Test the run_analysis entry point."""
    res_path = Path(tmpdir / "figures")

    # Run with CLI
    runner = CliRunner()
    runner.invoke(
        cli.cli,
        [
            "run_analysis",
            "--orig-path",
            str(single_pop_data_dir),
            "--diam-path",
            str(single_pop_diametrized_data_dir),
            "--mtypes-file",
            str(single_pop_data_dir / "neurondb.dat"),
            "--out-dir",
            str(res_path),
            "--cumulative",
            "--individual",
            "--violin",
            "--ext",
            ".pdf",
        ],
        catch_exceptions=False,
    )

    # Check results
    assert sorted(
        [i.relative_to(res_path).as_posix() for i in res_path.glob("**") if i != res_path]
    ) == [
        "apical",
        "apical/L5_TPC_A_cumulative_section_path_distances_areas_individual",
        "apical/L5_TPC_A_cumulative_section_path_distances_volumes_individual",
        "axon",
        "axon/L5_TPC_A_cumulative_section_path_distances_areas_individual",
        "axon/L5_TPC_A_cumulative_section_path_distances_volumes_individual",
        "basal",
        "basal/L5_TPC_A_cumulative_section_path_distances_areas_individual",
        "basal/L5_TPC_A_cumulative_section_path_distances_volumes_individual",
    ]

    # Test with neither cumukative nor violin
    with pytest.raises(ValueError):
        runner.invoke(
            cli.cli,
            [
                "run_analysis",
                "--orig-path",
                str(single_pop_data_dir),
                "--diam-path",
                str(single_pop_diametrized_data_dir),
                "--mtypes-file",
                str(single_pop_data_dir / "neurondb.dat"),
                "--out-dir",
                str(res_path),
                "--ext",
                ".pdf",
            ],
            catch_exceptions=False,
        )



"""Test the build_diameters module."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

import morphio.mut
import numpy as np
import pytest
from morphio import PointLevel
from morphio import SectionType

from diameter_synthesis import build_diameters
from diameter_synthesis.exception import DiameterSynthesisError

from .testing_tools import _compare_diameters


def test_build_simple(config, model_params, small_morph):
    """Test the main build function to diametrize a neuron."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"

    build_diameters.build(small_morph, neurite_types, model_params[model][mtype], config[model])

    expected = morphio.mut.Morphology(small_morph)
    diameters = {
        0: [1.2997824, 0.9000233, 0.72673666],  # basal
        1: [1.6028887, 1.2359781, 0.91419667],  # basal
        2: [0.7737082, 0.6514245, 0.61029184],  # basal
        3: [0.727182, 0.5553919, 0.540321],  # basal
        4: [3.2547958, 2.7254605, 2.1961255],  # apical
        5: [1.2261434, 0.5947401, 0.44064492],  # apical
        6: [1.2281191, 1.0033262, 0.85901225],  # apical
        7: [0.5838221, 0.40629458, 0.38336557],  # apical
        8: [0.6178318, 0.6467553, 0.70707476],  # apical
        9: [0.58992374, 0.4161563, 0.4003501],  # apical
        10: [0.5538684, 0.47851545, 0.4900443],  # apical
        11: [0.44196147, 0.42748648, 0.42748648],  # apical
        12: [0.43044534, 0.38146895, 0.39314994],  # apical
        13: [0.2, 0.2, 0.2],  # axon
        14: [0.2, 0.2, 0.2],  # axon
        15: [0.2, 0.2, 0.2],  # basal
        16: [0.2, 0.2, 0.2],  # basal
    }
    for section_id, section in expected.sections.items():
        section.diameters = diameters[section_id]

    _compare_diameters(expected, small_morph)


def test_build_simple_with_apical_sections(config, model_params, small_morph):
    """Test the main build function to diametrize a neuron with a known apical section."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"
    model_params[model][mtype]["apical_point_sec_ids"] = [6]

    build_diameters.build(small_morph, neurite_types, model_params[model][mtype], config[model])

    expected = morphio.mut.Morphology(small_morph)
    diameters = {
        0: [1.6146091, 1.1505961, 0.74188745],  # basal
        1: [1.8352976, 1.5289228, 1.2671413],  # basal
        2: [0.96662533, 0.72055197, 0.6243407],  # basal
        3: [0.894905, 0.653613, 0.56571424],  # basal
        4: [3.5404124, 2.9960587, 2.4517055],  # apical
        5: [0.4869547, 0.45853767, 0.4576969],  # apical
        6: [2.4517055, 2.073075, 1.769174],  # apical
        7: [0.8419878, 0.60131085, 0.44629398],  # apical
        8: [1.3291866, 1.1325629, 0.979848],  # apical
        9: [0.59224963, 0.48624295, 0.4726326],  # apical
        10: [0.6087809, 0.511539, 0.4896373],  # apical
        11: [0.41421217, 0.3673572, 0.39478624],  # apical
        12: [0.43502575, 0.43636972, 0.4498969],  # apical
        13: [0.2, 0.2, 0.2],  # axon
        14: [0.2, 0.2, 0.2],  # axon
        15: [0.2, 0.2, 0.2],  # basal
        16: [0.2, 0.2, 0.2],  # basal
    }
    for section_id, section in expected.sections.items():
        section.diameters = diameters[section_id]

    _compare_diameters(expected, small_morph)


def test_build_simple_with_several_apical_sections(config, model_params, small_morph):
    """Test the main build function to diametrize a neuron with a known apical section."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"

    # Duplicate the apical
    for section in small_morph.sections[4].iter():
        if section.id != 4:
            append_fun = small_morph.sections[section.parent.id + 13].append_section
        else:
            append_fun = small_morph.append_root_section
        pts = section.points.copy()
        pts[:, 0] += 1000
        append_fun(PointLevel(pts, section.diameters.copy()), SectionType(3))

    model_params[model][mtype]["apical_point_sec_ids"] = [6, 19]

    build_diameters.build(small_morph, neurite_types, model_params[model][mtype], config[model])

    expected = morphio.mut.Morphology(small_morph)
    diameters = {
        0: [1.6581707, 1.0526989, 0.6265956],  # basal
        1: [1.8037863, 1.4524208, 1.1289493],  # basal
        2: [0.7909685, 0.64985496, 0.56315005],  # basal
        3: [0.86445045, 0.6507469, 0.56271935],  # basal
        4: [3.3207638, 2.783594, 2.294971],  # apical
        5: [0.54798895, 0.4854589, 0.48255223],  # apical
        6: [2.294971, 1.9488478, 1.6358488],  # apical
        7: [0.47745314, 0.44113955, 0.48057836],  # apical
        8: [1.6358488, 1.5305059, 1.4305559],  # apical
        9: [0.86752427, 0.6409085, 0.47136444],  # apical
        10: [0.9211162, 0.71516794, 0.56216586],  # apical
        11: [0.43250427, 0.41028976, 0.4410184],  # apical
        12: [0.45779485, 0.42024216, 0.39077753],  # apical
        13: [0.2, 0.2, 0.2],  # axon
        14: [0.2, 0.2, 0.2],  # axon
        15: [0.2, 0.2, 0.2],  # basal
        16: [0.2, 0.2, 0.2],  # basal
        17: [1.2681566, 0.8854807, 0.66164625],  # apical
        18: [0.5188144, 0.45719615, 0.45719615],  # apical
        19: [0.66164625, 0.59810007, 0.6117419],  # apical
        20: [0.6117419, 0.46071988, 0.46071988],  # apical
        21: [0.4938531, 0.46393174, 0.49159345],  # apical
        22: [0.42158556, 0.41690236, 0.41896287],  # apical
        23: [0.41675425, 0.4613071, 0.49716002],  # apical
        24: [0.41612378, 0.4461469, 0.46464235],  # apical
        25: [0.4088277, 0.4126462, 0.42915756],  # apical
    }
    for section_id, section in expected.sections.items():
        section.diameters = diameters[section_id]

    _compare_diameters(expected, small_morph)


def test_build(config, model_params, neuron, neuron_diametrized):
    """Test the main build function to diametrize a neuron."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"

    build_diameters.build(neuron, neurite_types, model_params[model][mtype], config[model])

    _compare_diameters(neuron_diametrized, neuron)


def test_build_no_seed(config, model_params, neuron, neuron_diametrized):
    """Test main build without seed in config."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"

    seed = config[model].pop("seed")
    rng = np.random.default_rng(seed)

    build_diameters.build(neuron, neurite_types, model_params[model][mtype], config[model], rng)

    _compare_diameters(neuron_diametrized, neuron)


def test_build_multiple_models_warning(config, model_params, neuron, neuron_diametrized, caplog):
    """Test main build with multiple models in config."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"
    config[model]["models"].append("SKIPPED MODEL")

    caplog.clear()
    caplog.set_level(logging.WARNING)
    build_diameters.build(neuron, neurite_types, model_params[model][mtype], config[model])

    module, level, entry = caplog.record_tuples[0]
    assert module == "diameter_synthesis.build_diameters"
    assert level == 30
    assert entry == "Several models provided, we will only use the first"

    _compare_diameters(neuron_diametrized, neuron)


def test_build_one_sample(test_data_path, config, model_params, neuron):
    """Test main build with 1 sample in config."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"
    config[model]["n_samples"] = 1

    build_diameters.build(neuron, neurite_types, model_params[model][mtype], config[model])

    neuron_diametrized = morphio.mut.Morphology(
        test_data_path / "C030796A-P3_lite_diametrized_1_sample.h5"
    )

    _compare_diameters(neuron_diametrized, neuron)


def test_build_no_sequential(test_data_path, config, model_params, neuron):
    """Test main build with no sequential in config."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "astrocyte"

    config[model] = config["generic"]
    config[model]["models"] = [model]
    config[model]["n_samples"] = 1
    model_params[model] = model_params["generic"]
    model_params[model][mtype]["sibling_ratios"]["apical"]["sequential"] = None

    build_diameters.build(neuron, neurite_types, model_params[model][mtype], config[model])

    neuron_diametrized = morphio.mut.Morphology(
        test_data_path / "C030796A-P3_lite_diametrized_astrocyte.h5"
    )

    _compare_diameters(neuron_diametrized, neuron)


def test_build_small_n_tries_warning(config, model_params, neuron, caplog):
    """Test main build with trunk_max_tries = 1."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"
    config[model]["trunk_max_tries"] = 1

    caplog.clear()
    caplog.set_level(logging.WARNING)
    build_diameters.build(neuron, neurite_types, model_params[model][mtype], config[model])

    assert len(caplog.record_tuples) == 8
    neurite_types = [
        "basal",
        "apical",
        "apical",
        "basal",
        "apical",
        "apical",
        "apical",
        "apical",
    ]
    for i in zip(neurite_types, caplog.record_tuples):
        neurite_type, (module, level, entry) = i
        assert module == "diameter_synthesis.build_diameters"
        assert level == 30
        assert entry == f"max tries attained with {neurite_type}"


def test_build_small_trunk_diam_warning(config, model_params, neuron, caplog):
    """Test trunk diam warning in main build."""
    neurite_types = ["basal", "apical"]
    mtype = "L5_TPC:A"
    model = "generic"

    TRUNK_FRAC_DECREASE = build_diameters.TRUNK_FRAC_DECREASE
    try:
        build_diameters.TRUNK_FRAC_DECREASE = 999
        caplog.clear()
        caplog.set_level(logging.WARNING)
        build_diameters.build(neuron, neurite_types, model_params[model][mtype], config[model])
    finally:
        build_diameters.TRUNK_FRAC_DECREASE = TRUNK_FRAC_DECREASE

    assert len(caplog.record_tuples) == 8
    for i in caplog.record_tuples:
        module, level, entry = i
        assert module == "diameter_synthesis.build_diameters"
        assert level == 30
        assert entry == "sampled trunk diameter < 0.01, so use 1 instead"


def test_select_model():
    """Test the _select_model function."""
    # pylint: disable=protected-access
    # pylint: disable=no-member
    f_generic = build_diameters._select_model("generic")
    assert f_generic.args == (
        {
            "mode_sibling": "threshold",
            "mode_diameter_power_relation": "threshold",
            "with_asymmetry": True,
            "reduction_factor_max": 1.0,
        },
    )

    f_astrocyte = build_diameters._select_model("astrocyte")
    assert f_astrocyte.args == (
        {
            "mode_sibling": "generic",
            "mode_diameter_power_relation": "generic",
            "with_asymmetry": True,
            "reduction_factor_max": 3.0,
        },
    )

    with pytest.raises(DiameterSynthesisError):
        build_diameters._select_model("UNKNOWN")


def test_sample_sibling_ratio(model_params):
    """Test the _sample_sibling_ratio function."""
    # pylint: disable=protected-access
    mtype = "L5_TPC:A"
    model = "generic"
    d_generic = build_diameters._sample_sibling_ratio(
        model_params[model][mtype], "basal", False, mode="generic"
    )
    assert d_generic == pytest.approx(0.8253851329090136)

    d_threshold = build_diameters._sample_sibling_ratio(
        model_params[model][mtype], "basal", False, mode="threshold"
    )
    assert d_threshold == pytest.approx(0.7244487906052952)

    with pytest.raises(DiameterSynthesisError):
        build_diameters._sample_sibling_ratio(
            model_params[model][mtype], "basal", False, mode="UNKNOWN"
        )


def test_sample_diameter_power_relation(model_params):
    """Test the _sample_diameter_power_relation function."""
    # pylint: disable=protected-access
    mtype = "L5_TPC:A"
    model = "generic"
    d_generic = build_diameters._sample_diameter_power_relation(
        model_params[model][mtype], "basal", False, mode="generic"
    )
    assert d_generic == pytest.approx(4.942498757112429)

    d_threshold = build_diameters._sample_diameter_power_relation(
        model_params[model][mtype], "basal", False, mode="threshold"
    )
    assert d_threshold == pytest.approx(5.494060358805385)

    d_threshold = build_diameters._sample_diameter_power_relation(
        model_params[model][mtype], "basal", False, mode="exact"
    )
    assert d_threshold == pytest.approx(1)

    with pytest.raises(DiameterSynthesisError):
        build_diameters._sample_diameter_power_relation(
            model_params[model][mtype], "basal", False, mode="UNKNOWN"
        )


def test_sample_daughter_diameters(neuron, model_params):
    """Test the _sample_diameter_power_relation function."""
    # pylint: disable=protected-access
    mtype = "L5_TPC:A"
    model = "generic"

    param_tree = {
        "mode_sibling": "threshold",
        "mode_diameter_power_relation": "threshold",
        "with_asymmetry": True,
        "reduction_factor_max": 1.0,
        "neurite_type": "basal",
        "asymmetry_threshold": 1.0,
        "trunk_diam": 2.131444808766456,
        "tot_length": 1085.1178037524223,
        "terminal_diam": 0.35602967635703164,
        "major_sections": [],
    }
    d = build_diameters._sample_daughter_diameters(
        neuron.sections[0], model_params[model][mtype], param_tree
    )
    assert d == pytest.approx([0.9373572556376409, 0.7736807430277025])

    param_tree["with_asymmetry"] = False
    d_no_asymmetry = build_diameters._sample_daughter_diameters(
        neuron.sections[0], model_params[model][mtype], param_tree
    )
    assert d_no_asymmetry == pytest.approx([1.120733059071045, 0.8804980886414063])


def test_diametrize_axon(small_morph):
    """Test the axon diametrizer."""
    build_diameters.diametrize_axon(small_morph)

    expected = morphio.mut.Morphology(small_morph)
    diameters = {
        0: [0.2, 0.2, 0.2],  # basal
        1: [0.2, 0.2, 0.2],  # basal
        2: [0.2, 0.2, 0.2],  # basal
        3: [0.2, 0.2, 0.2],  # basal
        4: [0.2, 0.2, 0.2],  # apical
        5: [0.2, 0.2, 0.2],  # apical
        6: [0.2, 0.2, 0.2],  # apical
        7: [0.2, 0.2, 0.2],  # apical
        8: [0.2, 0.2, 0.2],  # apical
        9: [0.2, 0.2, 0.2],  # apical
        10: [0.2, 0.2, 0.2],  # apical
        11: [0.2, 0.2, 0.2],  # apical
        12: [0.2, 0.2, 0.2],  # apical
        13: [0.2, 0.2, 0.9],  # axon
        14: [0.2, 0.2, 0.9],  # axon
        15: [0.376628, 0.326628, 0.276628],  # basal
        16: [0.753256, 0.703256, 0.653256],  # basal
    }
    for section_id, section in expected.sections.items():
        section.diameters = diameters[section_id]

    _compare_diameters(expected, small_morph, rtol=1e-6)



"""Functions to fit distributions to parameters of diameter models."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=import-outside-toplevel
import logging

import numpy as np

from diameter_synthesis.exception import DiameterSynthesisError

MIN_DATA_POINTS = 1  # minimum number of points to fit a distribution
A_MAX = 4
A_MIN = 0.3

N_BINS = 10
PERCENTILE = 5
MIN_SAMPLE_NUM = 10

MAX_TRUNCATE_TRIES = 100

L = logging.getLogger(__name__)
np.seterr(invalid="ignore", divide="ignore")


def _truncate(sample_func, min_value, max_value):
    """Ensure sample is within bounds."""
    sample = sample_func()
    n_tries = 0
    if min_value == max_value:
        return min_value
    while sample > max_value or sample < min_value:
        sample = sample_func()
        n_tries += 1
        if n_tries >= MAX_TRUNCATE_TRIES:
            raise DiameterSynthesisError(
                f"Could not truncate the sample between {min_value} and {max_value} "
                f"(the last value was {sample})"
            )
    return sample


def fit_distribution(all_data, distribution, attribute_name=None, extra_params=None):
    """Fit a distribution from data.

    Args:
        all_data (list/array): list of data points to fit a distribution to.
        distribution (str): Distribution name.
        attribute_name (str): Name of additional attribute to fit.
        extra_params (dict): Possible additional parameters for the fit.

    Returns:
        dict: parameters of the fit.
    """
    if attribute_name is not None:
        if extra_params is None:
            raise DiameterSynthesisError(
                "'extra_params' can not be None when 'attribute_name' is not None"
            )
        threshold = (
            extra_params.get("features", {})
            .get(attribute_name, {})
            .get(extra_params["neurite_type"])
        )
        if threshold is None:
            raise DiameterSynthesisError(
                "The threshold could not be retrieved from the config in 'features'->"
                f"'{attribute_name}'->'{extra_params['neurite_type']}'"
            )
        attribute = np.asarray(all_data, dtype=np.float32)[:, 1]
        data = np.asarray(all_data, dtype=np.float32)[:, 0]
        data = data[attribute < threshold]
    else:
        data = all_data

    if len(data) < MIN_DATA_POINTS:
        L.warning(
            "Not enough data to fit distribution %s with %s points",
            extra_params["name"],
            len(data),
        )
        return {
            "a": 0.0,
            "loc": 0.0,
            "scale": 0.0,
            "min": 0.0,
            "max": 0.1,
            "num_value": len(data),
        }

    if distribution == "expon_rev":
        from scipy.stats import expon

        loc, scale = expon.fit(-np.array(data))
        return {
            "loc": float(loc),
            "scale": float(scale),
            "min": float(np.percentile(data, PERCENTILE)),
            "max": float(max(data)),
            "num_value": float(len(data)),
        }

    if distribution == "exponnorm":
        from scipy.stats import exponnorm

        var_a, loc, scale = exponnorm.fit(data)
        if var_a > A_MAX:
            var_a, loc, scale = exponnorm.fit(data, f0=A_MAX)
        if var_a < A_MIN:
            var_a, loc, scale = exponnorm.fit(data, f0=A_MIN)
        return {
            "a": float(var_a),
            "loc": float(loc),
            "scale": float(scale),
            "min": float(np.percentile(data, PERCENTILE)),
            "max": float(np.percentile(data, 100 - PERCENTILE)),
            "num_value": int(len(data)),
        }

    if distribution == "gamma":
        from scipy.stats import gamma

        var_a, loc, scale = gamma.fit(data, floc=-1e-9)
        if var_a > A_MAX:
            var_a, loc, scale = gamma.fit(data, f0=A_MAX, floc=-1e-9)
        if var_a < A_MIN:
            var_a, loc, scale = gamma.fit(data, f0=A_MIN, floc=-1e-9)
        return {
            "a": float(var_a),
            "loc": float(loc),
            "scale": float(scale),
            "min": float(np.percentile(data, PERCENTILE)),
            "max": float(np.percentile(data, 100 - PERCENTILE)),
            "num_value": int(len(data)),
        }

    raise DiameterSynthesisError("Distribution not understood")


def sample_distribution(model, rng=np.random):
    """Sample from a distribution.

    Args:
        model (dict): the model to use.
        rng (numpy.random.Generator): the random number generator.

    Returns:
        float: the value of the distribution at the given position.
    """
    if "a" in model["params"]:
        a_clip = np.clip(model["params"]["a"], A_MIN, A_MAX)

    if model["distribution"] == "constant":
        return model["params"]["value"]

    if model["distribution"] == "expon_rev":
        return _truncate(
            lambda: -(
                model["params"]["loc"] + model["params"]["scale"] * rng.standard_exponential()
            ),
            model["params"]["min"],
            model["params"]["max"],
        )

    if model["distribution"] == "exponnorm":
        return _truncate(
            lambda: (
                model["params"]["loc"]
                + model["params"]["scale"]
                * (rng.standard_exponential() * a_clip + rng.standard_normal())
            ),
            model["params"]["min"],
            model["params"]["max"],
        )

    if model["distribution"] == "gamma":
        return _truncate(
            lambda: model["params"]["loc"] + model["params"]["scale"] * rng.standard_gamma(a_clip),
            model["params"]["min"],
            model["params"]["max"],
        )

    raise DiameterSynthesisError("Distribution not understood")


def evaluate_distribution(value, distribution, params):
    """Evaluate the fit of a distribution.

    Args:
        value (float): the position at which the distribution is evaluated.
        distribution (str): the name of the distribution.
        params (dict): the parameters of the distributions.

    Returns:
        float: the value of the distribution at the given position.
    """
    if distribution == "expon_rev":
        from scipy.stats import expon

        return expon.pdf(-value, params["loc"], params["scale"])

    if distribution == "exponnorm":
        from scipy.stats import exponnorm

        return exponnorm.pdf(value, params["a"], params["loc"], params["scale"])

    if distribution == "gamma":
        from scipy.stats import gamma

        return gamma.pdf(value, params["a"], params["loc"], params["scale"])

    raise DiameterSynthesisError("Distribution not understood")



"""Test the morph_functions module."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diameter_synthesis import morph_functions
from diameter_synthesis.exception import DiameterSynthesisError


def test_partition_asymmetry_length(neuron):
    """Test the partition_asymmetry_length function."""
    section = neuron.root_sections[0]
    res = morph_functions.partition_asymmetry_length(section)
    assert res == pytest.approx(157.6603)

    # Test with only one child section
    one_child_section = neuron.root_sections[1]
    neuron.delete_section(one_child_section.children[0], recursive=True)
    with pytest.raises(DiameterSynthesisError):
        morph_functions.partition_asymmetry_length(one_child_section)


def test_compute_sibling_ratios(single_neurite):
    """Test the compute_sibling_ratios function."""
    res = morph_functions.compute_sibling_ratios(single_neurite, "mean")
    assert res == pytest.approx([0.508197, 0.434402, 1, 1, 0.395023, 1, 1])

    res_bounds = morph_functions.compute_sibling_ratios(
        single_neurite, "mean", bounds=[0, 1 + 1e-5]
    )
    assert res_bounds == pytest.approx(res)


def test_compute_diameter_power_relation(single_neurite):
    """Test the compute_diameter_power_relation function."""
    res = morph_functions.compute_diameter_power_relation(single_neurite, "mean")
    assert res == pytest.approx(
        [6.964765, 4.492706, 5.942041, 2.534333, 7.254541, 2.822203, 16.933426]
    )

    res_bounds = morph_functions.compute_diameter_power_relation(
        single_neurite, "mean", bounds=[0, 100]
    )
    assert res_bounds == pytest.approx(res)


def test_terminal_diameters(single_neurite):
    """Test the terminal_diameters function."""
    res = morph_functions.terminal_diameters(single_neurite, "mean")
    assert res == pytest.approx([0.3, 0.3, 0.23, 0.23, 0.3, 0.3, 0.23, 0.23])

    res_bounds = morph_functions.terminal_diameters(single_neurite, "mean", bounds=[0, 100])
    assert res_bounds == pytest.approx(res)

    res_first = morph_functions.terminal_diameters(single_neurite, "first")
    assert res_first == pytest.approx(res)

    with pytest.raises(DiameterSynthesisError):
        morph_functions.terminal_diameters(single_neurite, "UNKNOWN")


def test_min_diameter(single_neurite):
    """Test the min_diameter function."""
    res = morph_functions.min_diameter(single_neurite)
    assert res == pytest.approx([0.23])

    res_bounds = morph_functions.min_diameter(single_neurite, bounds=[0, 100])
    assert res_bounds == pytest.approx(res)


def test_max_diameter(single_neurite):
    """Test the max_diameter function."""
    res = morph_functions.max_diameter(single_neurite)
    assert res == pytest.approx([1.84])

    res_bounds = morph_functions.max_diameter(single_neurite, bounds=[0, 100])
    assert res_bounds == pytest.approx(res)


def test_trunk_diameter(single_neurite):
    """Test the trunk_diameter function."""
    res = morph_functions.trunk_diameter(single_neurite)
    assert res == pytest.approx([1.84])

    res_bounds = morph_functions.trunk_diameter(single_neurite, bounds=[0, 100])
    assert res_bounds == pytest.approx(res)

    res_bounds = morph_functions.trunk_diameter(single_neurite, method="mean")
    assert res_bounds == pytest.approx(res)

    res_bounds = morph_functions.trunk_diameter(single_neurite, method="first")
    assert res_bounds == pytest.approx(res)


def test_taper(single_neurite):
    """Test the taper function."""
    params = {"min": -1e-9, "max": 1e-9}
    res = morph_functions.taper(single_neurite, params)
    assert res == pytest.approx([0.0] * 12)

    res_tot_length = morph_functions.taper(single_neurite, params, "tot_length")
    assert len(res_tot_length) == 1
    assert res_tot_length[0] == pytest.approx([0.0, 828.830532])

    # Test with None params
    with pytest.raises(DiameterSynthesisError):
        morph_functions.taper(single_neurite, None)


def test_get_additional_attribute(single_neurite):
    """Test the get_additional_attribute function."""
    res_asymmetry = morph_functions.get_additional_attribute("asymmetry", single_neurite)
    assert res_asymmetry == pytest.approx(
        [0.19022014, 0.02628145, 0.16578421, 0.16073337, 0.00042449, 0.00093826, 0.02613346],
        rel=1e-5,
    )

    res_asymmetry_threshold = morph_functions.get_additional_attribute(
        "asymmetry_threshold", single_neurite
    )
    assert res_asymmetry_threshold == pytest.approx(res_asymmetry)

    section = next(single_neurite.iter_sections())
    res_asymmetry_pair = morph_functions.get_additional_attribute("asymmetry_pair", section=section)
    assert len(res_asymmetry_pair) == 1
    assert res_asymmetry_pair[0] == pytest.approx([7, 7])

    res_tot_length = morph_functions.get_additional_attribute("tot_length", single_neurite)
    assert res_tot_length == pytest.approx([828.83])

    res_max_path = morph_functions.get_additional_attribute("max_path", single_neurite)
    assert res_max_path == pytest.approx([325.016])

    res_max_branch = morph_functions.get_additional_attribute("max_branch", single_neurite)
    assert res_max_branch == pytest.approx([3])

    res_root_strahler = morph_functions.get_additional_attribute("root_strahler", single_neurite)
    assert res_root_strahler == pytest.approx([4])

    res_sibling = morph_functions.get_additional_attribute("sibling", single_neurite)
    assert res_sibling == pytest.approx([0.508197, 0.434402, 1, 1, 0.395023, 1, 1])

    res_sibling_section = morph_functions.get_additional_attribute("sibling", section=section)
    assert res_sibling_section == pytest.approx(0.508197)

    # Test with unknown attribute name
    with pytest.raises(DiameterSynthesisError):
        morph_functions.get_additional_attribute("UNKNOWN", single_neurite)

    # Test with no neurite nor section
    with pytest.raises(DiameterSynthesisError):
        morph_functions.get_additional_attribute("asymmetry")

    # Test with both neurite and section
    for attr in [
        "asymmetry",
        "asymmetry_threshold",
        "asymmetry_pair",
        "tot_length",
        "max_path",
        "max_branch",
        "root_strahler",
        "sibling",
    ]:
        with pytest.raises(DiameterSynthesisError):
            morph_functions.get_additional_attribute(attr)


def test_add_additional_attributes(single_neurite):
    """Test the add_additional_attributes function."""
    res_max_branch = morph_functions.add_additional_attributes(
        [-1, 1, 10], single_neurite, "max_branch"
    )
    assert len(res_max_branch) == 1
    assert res_max_branch[0] == pytest.approx([-1, 3])

    res_max_branch_bounds = morph_functions.add_additional_attributes(
        [-1, 1, 10], single_neurite, "max_branch", bounds=[-1000, 1000]
    )
    assert len(res_max_branch_bounds) == 1
    assert res_max_branch_bounds[0] == pytest.approx([-1, 3])



"""Plotting functions."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import multiprocessing
import os
import re
from copy import copy
from functools import partial
from pathlib import Path

import matplotlib
import matplotlib.pyplot as plt
import neurom as nm
import numpy as np
import pandas
from matplotlib.backends.backend_pdf import PdfPages
from neurom import APICAL_DENDRITE
from neurom import AXON
from neurom import BASAL_DENDRITE
from neurom import COLS
from neurom import get
from neurom import iter_sections
from neurom.geom import bounding_box
from neurom.view import matplotlib_impl
from scipy import stats
from tqdm import tqdm

from diameter_synthesis import utils
from diameter_synthesis.distribution_fitting import evaluate_distribution

# pylint: disable=too-many-statements,too-many-locals,too-many-arguments

matplotlib.use("Agg")
L = logging.getLogger(__name__)
COLORS = {"basal": "r", "apical": "m", "axon": "b"}

CUMULATIVE_FEATURE_PAIRS = [
    ("section_path_distances", "section_volumes"),
    ("section_path_distances", "section_areas"),
]

VIOLIN_FEATURES_LIST = ["segment_radii", "section_areas", "section_volumes"]

VIOLIN_FEATURES_NAME = ["Segment radii", "Section areas", "Section volumes"]

NEURITE_STR_TO_TYPES = {
    "basal": nm.NeuriteType.basal_dendrite,
    "apical": nm.NeuriteType.apical_dendrite,
    "axon": nm.NeuriteType.axon,
}


def _compute_neurite_diff(neuron_orig, neuron_new, neuron_diff_pos, neuron_diff_neg, neurite_types):
    """Compute the differences between neurite diameters."""
    for neurite_type in neurite_types:
        neurites_orig = [
            neurite
            for neurite in neuron_orig.neurites
            if neurite.type == NEURITE_STR_TO_TYPES[neurite_type]
        ]
        neurites_new = [
            neurite
            for neurite in neuron_new.neurites
            if neurite.type == NEURITE_STR_TO_TYPES[neurite_type]
        ]
        neurites_diff_neg = [
            neurite
            for neurite in neuron_diff_neg.neurites
            if neurite.type == NEURITE_STR_TO_TYPES[neurite_type]
        ]
        neurites_diff_pos = [
            neurite
            for neurite in neuron_diff_pos.neurites
            if neurite.type == NEURITE_STR_TO_TYPES[neurite_type]
        ]

        for neurite_orig, neurite_new, neurite_diff_pos, neurite_diff_neg in zip(
            neurites_orig, neurites_new, neurites_diff_pos, neurites_diff_neg
        ):
            orig = []
            new = []
            for section_orig, section_new in zip(
                iter_sections(neurite_orig), iter_sections(neurite_new)
            ):
                orig.append(section_orig.points)
                new.append(section_new.points)

            for j, section in enumerate(iter_sections(neurite_diff_pos)):
                _diff = new[j][:, COLS.R] - orig[j][:, COLS.R]
                _diff[_diff < 0] = 0
                diff = copy(orig[j])
                diff[:, COLS.R] = _diff
                section.points = diff

            for j, section in enumerate(iter_sections(neurite_diff_neg)):
                _diff = -(new[j][:, COLS.R] - orig[j][:, COLS.R])
                _diff[_diff < 0] = 0
                diff = copy(orig[j])
                diff[:, COLS.R] = _diff
                section.points = diff


def plot_diameter_diff(neuron_name, neuron_new, neurite_types, folder, ext=".png"):
    """Plot original morphology, new one and differences."""
    if not Path(folder).exists():
        os.mkdir(folder)

    neuron_orig = nm.load_morphology(neuron_name)
    neuron_diff_pos = nm.load_morphology(neuron_name)
    neuron_diff_neg = nm.load_morphology(neuron_name)
    _compute_neurite_diff(neuron_orig, neuron_new, neuron_diff_pos, neuron_diff_neg, neurite_types)

    bbox = bounding_box(neuron_orig)

    fig, axs = plt.subplots(2, 2, figsize=(10, 10))

    matplotlib_impl.plot_morph(neuron_orig, axs[0, 0])
    axs[0, 0].set_title("Original neuron")
    axs[0, 0].set_xlim([bbox[0, 0], bbox[1, 0]])
    axs[0, 0].set_ylim([bbox[0, 1], bbox[1, 1]])
    axs[0, 0].set_aspect("equal")

    matplotlib_impl.plot_morph(neuron_new, axs[0, 1])
    axs[0, 1].set_title("New neuron")
    axs[0, 1].set_xlim([bbox[0, 0], bbox[1, 0]])
    axs[0, 1].set_ylim([bbox[0, 1], bbox[1, 1]])
    axs[0, 1].set_aspect("equal")

    matplotlib_impl.plot_morph(neuron_diff_pos, axs[1, 0])
    axs[1, 0].set_title("Positive diameter differences")
    axs[1, 0].set_xlim([bbox[0, 0], bbox[1, 0]])
    axs[1, 0].set_ylim([bbox[0, 1], bbox[1, 1]])
    axs[1, 0].set_aspect("equal")

    matplotlib_impl.plot_morph(neuron_diff_neg, axs[1, 1])
    axs[1, 1].set_title("Negative diameter differences")
    axs[1, 1].set_xlim([bbox[0, 0], bbox[1, 0]])
    axs[1, 1].set_ylim([bbox[0, 1], bbox[1, 1]])
    axs[1, 1].set_aspect("equal")

    fig.savefig((Path(folder) / neuron_name.stem).with_suffix(ext), dpi=500)
    plt.close()


def _plot_attribute_scatter(data, model, neurite_types, fig_name, figsize, ext):
    """Plot scatter of parameters with attribute if any attributes."""
    plt.figure(figsize=figsize)
    save_plot = False
    max_val = -1e10
    min_val = 1e10
    for neurite_type in neurite_types:
        if len(model[neurite_type]) > 0:
            if model[neurite_type]["sequential"] == "asymmetry_threshold":
                save_plot = True
                tpes = np.asarray(data[neurite_type])[:, 1]
                values = np.asarray(data[neurite_type])[:, 0]
                plt.scatter(values, tpes, s=5, c=COLORS[neurite_type], alpha=0.5)
                plt.axhline(0.2, c="k")

                min_val = min(min_val, model[neurite_type]["params"]["min"])
                max_val = max(max_val, model[neurite_type]["params"]["max"])

    if Path(fig_name).name == "sibling_ratios":
        plt.gca().set_xlim(0.0, 1.0)
    else:
        plt.gca().set_xlim(min_val * 0.5, max_val * 2.0)

    if save_plot:
        plt.savefig(str(fig_name) + "_scatter" + ext)

    plt.close()


def plot_distribution_fit(data, model, neurite_types, fig_name="test", ext=".png", figsize=(5, 4)):
    """Plot the data distribution and its fit."""
    _plot_attribute_scatter(data, model, neurite_types, fig_name, figsize, ext)

    plt.figure()
    for neurite_type in neurite_types:
        if len(model[neurite_type]) > 0:
            min_val = model[neurite_type]["params"]["min"]
            max_val = model[neurite_type]["params"]["max"]
            if Path(fig_name).name in ["sibling_ratios", "tapers"]:
                hist_range = [min_val, max_val]
            else:
                hist_range = [min_val * 0.5, max_val * 2.0]

            if len(data[neurite_type]) > 0:
                if len(np.shape(data[neurite_type])) > 1:
                    data_hist = np.array(data[neurite_type])[:, 0]
                else:
                    data_hist = data[neurite_type]

                plt.hist(
                    data_hist,
                    bins=30,
                    density=True,
                    color=COLORS[neurite_type],
                    alpha=0.5,
                    range=hist_range,
                )

                values = np.linspace(min_val, max_val, 1000)
                plt.plot(
                    values,
                    evaluate_distribution(
                        values,
                        model[neurite_type]["distribution"],
                        model[neurite_type]["params"],
                    ),
                    c=COLORS[neurite_type],
                    lw=3,
                    ls="--",
                    label=neurite_type,
                )
                plt.legend(loc="best")
            else:
                L.warning("No data to plot")

    plt.savefig(str(fig_name) + ext, bbox_inches="tight")
    plt.close()


def _create_data(
    feature1, feature2, original_cells, diametrized_cells, step_size, neurite_types
):  # noqa, pylint: disable=too-many-locals,too-many-arguments
    def feature_data(cell, neurite_type):
        nm_neurite_type = NEURITE_STR_TO_TYPES[neurite_type]
        return [get(feat, cell, neurite_type=nm_neurite_type) for feat in (feature1, feature2)]

    def create_paired_features(cell_list1, cell_list2, neurite_type):
        for cell1, cell2 in zip(cell_list1, cell_list2):
            yield feature_data(cell1, neurite_type), feature_data(cell2, neurite_type)

    def create_bins(step_size, max_value):
        bins = np.arange(0.0, max_value + step_size, step_size)
        bin_centers = 0.5 * step_size + bins[:-1]
        return bin_centers, bins

    def find_upper_bound(pairs):
        return max(max(max(vals1), max(vals2)) for (vals1, _), (vals2, _) in pairs)

    def per_neurite_data(original_cells, diametrized_cells, neurite_types):
        for _, neurite_type in enumerate(neurite_types):
            data = list(create_paired_features(original_cells, diametrized_cells, neurite_type))
            if len(data[0][0][0]) > 0:
                yield data

    assert len(original_cells) == len(diametrized_cells)

    n_cells = len(original_cells)
    iter_neurite_data = per_neurite_data(original_cells, diametrized_cells, neurite_types)
    for _, data_pairs in enumerate(iter_neurite_data):

        try:
            upper_bound = find_upper_bound(data_pairs)
        except BaseException:  # pylint: disable=broad-except
            L.exception("failed to find upper bound, most likely due to no data points")
            upper_bound = 200

        bin_centers, bins = create_bins(step_size, upper_bound)

        stats1 = np.empty((n_cells, len(bin_centers)), dtype=float)
        stats2 = np.empty_like(stats1)

        for i, ((metric1, data1), (metric2, data2)) in enumerate(data_pairs):

            res1 = stats.binned_statistic(
                metric1, data1, statistic="sum", bins=bins, range=(0, upper_bound)
            )
            res2 = stats.binned_statistic(
                metric2, data2, statistic="sum", bins=bins, range=(0, upper_bound)
            )

            stats1[i] = np.cumsum(res1.statistic)
            stats2[i] = np.cumsum(res2.statistic)

        yield bin_centers, stats1, stats2


def plot_cumulative_distribution(
    original_cells,
    diametrized_cells,
    feature1,
    feature2,
    neurite_types,
    step_size=1.0,
    auto_limit=True,
):
    """Plot the cumulative distribution of features.

    It plots feature2 with respect to the metric values determined via feature1.

    Args:
        original_cells: list of NeuroM objects.
        diametrized_cells (list): The new cells with the changed diameters.
        feature1: the metric feature.
        feature2: the cumulative distribution feature.
        neurite_types (string): The list neurite types to be considered. e.g. ['basal', 'axon'].
        step_size (float): The step size of the cumulative histogram.
        auto_limit (bool): automatically compute limits.

    Examples of metric features (feature1):
        - segment_radial_distances
        - segment_path_distances (not implemented yet)

    Examples of cumulative distribution features (feature2):
        - segment_volumes
        - segment_surface_areas (not implemented yet)
    """
    assert len(original_cells) == len(diametrized_cells)

    data_generator = _create_data(
        feature1, feature2, original_cells, diametrized_cells, step_size, neurite_types
    )

    fig, axes = plt.subplots(3, 1, figsize=(5, 12))

    for (bin_centers, stats1, stats2) in data_generator:

        means = stats1.mean(axis=0)
        color = "C0"
        axes[0].plot(bin_centers, means, c=color, linestyle="-", lw=3, label="original cells")

        if len(stats1) > 1:
            for st1 in stats1:
                axes[0].plot(bin_centers, st1, c=color, linestyle="-", lw=0.5, alpha=0.2)

            sdevs = stats1.std(axis=0)
            axes[0].plot(bin_centers, means - sdevs, c=color, linestyle="--", lw=3)
            axes[0].plot(bin_centers, means + sdevs, c=color, linestyle="--", lw=3)

        means = stats2.mean(axis=0)

        color = "C1"
        axes[0].plot(
            bin_centers,
            means,
            c=color,
            linestyle="-",
            lw=3,
            label="rediametrized cells",
        )
        axes[0].legend(loc="best")
        if len(stats2) > 1:
            for st2 in stats2:
                axes[0].plot(bin_centers, st2, c=color, linestyle="-", lw=0.3, alpha=0.5)

            sdevs = stats2.std(axis=0)
            axes[0].plot(bin_centers, means - sdevs, c=color, linestyle="--", lw=3)
            axes[0].plot(bin_centers, means + sdevs, c=color, linestyle="--", lw=3)

        axes[0].set_xlabel("path distance")
        axes[0].set_ylabel("cumulative section areas")

        axes[0].set_xlabel("path distance")
        axes[0].set_ylabel("cumulative section areas")

        stats1[stats1 == 0] = 1
        diffs = stats1 - stats2  # / stats1

        diff_means = diffs.mean(axis=0)

        color = "C2"
        if len(diffs) > 1:
            for dfs in diffs:
                axes[1].plot(bin_centers, dfs, c=color, linestyle="-", lw=0.3, alpha=0.5)

            diff_sdevs = diffs.std(axis=0)
            axes[1].plot(bin_centers, diff_means - diff_sdevs, c=color, linestyle="--", lw=3)
            axes[1].plot(bin_centers, diff_means + diff_sdevs, c=color, linestyle="--", lw=3)

        axes[1].plot(bin_centers, diff_means, c=color, linestyle="-", lw=3)
        axes[1].axhline(0, ls="--", c="k")

        axes[1].set_xlabel("path distance")
        axes[1].set_ylabel("difference in cumulative section areas")

        if auto_limit:
            lim_min = 0.5 * np.min(stats1[:, -1])
            lim_max = np.max(stats1[:, -1])
        else:
            lim_min = 5000
            lim_max = 30000

        axes[2].scatter(stats1[:, -1], stats2[:, -1], c=color, marker="o", s=5)

        var_x = np.arange(lim_min, lim_max)

        axes[2].plot(var_x, var_x, ls="-", c="k")
        axes[2].set_xlim(lim_min, lim_max)
        axes[2].set_ylim(lim_min, lim_max)
        axes[2].set_title(
            "L2 error = " + str(np.around(np.linalg.norm(stats2[:, -1] - stats1[:, -1]), 1)),
            loc="left",
        )

        axes[2].set_xlabel("total surface area of original cells")
        axes[2].set_ylabel("total surface area of diametrized cells")

    return fig, axes


def _split_prefix(neurom_feature_name):
    """Split prefix."""
    name_list = neurom_feature_name.split("_")

    prefix = name_list[0]
    basename = "_".join(name_list[1:])

    return prefix, basename


def make_cumulative_figures(
    original_cells,
    diametrized_cells,
    feature1,
    feature2,
    neurite_types,
    out_dir,
    individual=False,
    figname_prefix="",
    ext=".png",
):
    """Make plots for cumulative distributions for a pair of features."""
    out_dir = Path(out_dir)
    out_dir.mkdir(parents=True, exist_ok=True)

    prefix1, basename1 = _split_prefix(feature1)
    prefix2, basename2 = _split_prefix(feature2)

    assert prefix1 == prefix2

    fig, _ = plot_cumulative_distribution(
        original_cells, diametrized_cells, feature1, feature2, neurite_types
    )

    if figname_prefix and not figname_prefix.endswith("_"):
        figname_prefix = figname_prefix + "_"
    figname_prefix = re.sub(r"[^\w\s-]", "_", figname_prefix)

    figure_name = f"{figname_prefix}cumulative_{prefix1}_{basename1}_{basename2}"
    fig.savefig(out_dir / f"{figure_name}{ext}", bbox_inches="tight")
    plt.close(fig)

    if individual:
        if not (out_dir / (figure_name + "_individual")).exists():
            os.mkdir(out_dir / (figure_name + "_individual"))

        for i, (original_cell, diametrized_cell) in enumerate(
            zip(original_cells, diametrized_cells)
        ):
            f, _ = plot_cumulative_distribution(
                [original_cell],
                [diametrized_cell],
                feature1,
                feature2,
                neurite_types,
                auto_limit=False,
            )
            fname = f"{figure_name}_{original_cell.name}{ext}"
            f.savefig(
                out_dir / (figure_name + "_individual") / (str(i) + "_" + fname),
                bbox_inches="tight",
            )
            plt.close(f)


def _load_morphologies(morph_path, mtypes_file="./neuronDB.xml"):
    """Load the morphologies from a directory, by mtypes or all at once."""
    morphologies_dict = utils.create_morphologies_dict(morph_path, mtypes_file=mtypes_file)
    return {
        mtype: [nm.load_morphology(i) for i in morphologies_dict[mtype]]
        for mtype in morphologies_dict
    }


def cumulative_analysis(
    original_path,
    diametrized_path,
    out_dir,
    individual=False,
    mtypes_file=None,
    neurite_types=None,
    ext=".png",
):
    """Make plots for cumulative distributions."""
    out_dir = Path(out_dir)
    out_dir.mkdir(parents=True, exist_ok=True)

    all_original_cells = _load_morphologies(original_path, mtypes_file=mtypes_file)
    all_diametrized_cells = _load_morphologies(diametrized_path, mtypes_file=mtypes_file)
    for mtype in tqdm(all_original_cells):
        original_cells = all_original_cells[mtype]
        diametrized_cells = all_diametrized_cells[mtype]
        for feature1, feature2 in CUMULATIVE_FEATURE_PAIRS:
            make_cumulative_figures(
                original_cells,
                diametrized_cells,
                feature1,
                feature2,
                neurite_types,
                out_dir,
                individual=individual,
                figname_prefix=mtype,
                ext=ext,
            )


def get_features_all(object1, object2, flist, neurite_type):
    """Compute features from module mod."""

    def _get_feature(feat, obj):
        v = get(feat, obj, neurite_type=neurite_type)
        return v if isinstance(v, list) else [v]

    collect_all = []
    for feat in flist:
        feature_pop = []
        for obj in object1:
            feature_pop = feature_pop + _get_feature(feat, obj)
        feature_neu = []
        for obj in object2:
            feature_neu = feature_neu + _get_feature(feat, obj)

        collect_all.append([feature_pop, feature_neu])
    return collect_all


def transform2DataFrame(data, pop_names, flist):
    """Return a DataFrame in the appropriate format from a set of features."""
    values = []
    names = []
    feat = []
    # Loop through each feature
    for i, d1 in enumerate(data):
        m = np.mean(d1[0])
        st = np.std(d1[0])
        # Loop through populations
        for j, d2 in enumerate(d1):
            values = values + [(d3 - m) / st for d3 in d2]
            names = names + len(d2) * [pop_names[j]]
            feat = feat + len(d2) * [flist[i]]

    frame = pandas.DataFrame({"Data": names, "Values": values, "Morphological features": feat})

    return frame


def plot_violins(data, x="Morphological features", y="Values", hues="Data", ax=None):
    """Plot the split violins of all features."""
    import seaborn  # pylint: disable=import-outside-toplevel

    if ax is None:
        fig = plt.figure(figsize=(12, 6))
        ax = plt.gca()
    else:
        fig = None

    seaborn.violinplot(
        x=x,
        y=y,
        hue=hues,
        data=data,
        palette="muted",
        split=True,
        inner="quartile",
        ax=ax,
    )
    ax.tick_params(axis="x", rotation=20)
    return fig, ax


def violin_analysis(
    original_path,
    diametrized_path,
    out_dir,
    mtypes_file=None,
    max_cells=200,
    with_axon=False,
):
    """Plot violin distributions."""
    out_dir = Path(out_dir)
    out_dir.mkdir(parents=True, exist_ok=True)

    orig_morphologies_dict = utils.create_morphologies_dict(original_path, mtypes_file=mtypes_file)
    diametrized_morphologies_dict = utils.create_morphologies_dict(
        diametrized_path, mtypes_file=mtypes_file
    )

    cells_data = [
        [orig_morphologies_dict[mtype], diametrized_morphologies_dict[mtype], mtype]
        for mtype in orig_morphologies_dict
    ]
    analyze_from_dict = partial(_analyze_from_dict, max_cells, with_axon=with_axon)

    pool = multiprocessing.Pool()  # pylint: disable=consider-using-with
    try:
        figs = list(
            tqdm(
                pool.imap_unordered(analyze_from_dict, cells_data),
                total=len(cells_data),
            )
        )
    finally:
        pool.close()
        pool.join()

    with PdfPages(out_dir / "morphometrics.pdf") as pdf:
        for mtype, fig in figs:
            if mtype is not None:
                pdf.savefig(fig)


def _analyze_from_dict(max_cells, cells, with_axon=False):
    cell_orig, cell_diametrized, mtype = cells
    cell_diametrized = cell_diametrized[:max_cells]
    cell_orig = cell_orig[:max_cells]
    original_cells = [nm.load_morphology(i) for i in cell_orig]
    diametrized_cells = [nm.load_morphology(i) for i in cell_diametrized]

    pop_names = ["Original cells of " + mtype, "Synthetised cells of " + mtype]
    data = get_features_all(
        original_cells,
        diametrized_cells,
        flist=VIOLIN_FEATURES_LIST,
        neurite_type=BASAL_DENDRITE,
    )

    data_frame = transform2DataFrame(data, pop_names, flist=VIOLIN_FEATURES_NAME)

    if with_axon:
        fig, axes = plt.subplots(nrows=3, ncols=1, figsize=(15, 10))
    else:
        fig, axes = plt.subplots(nrows=2, ncols=1, figsize=(15, 10))
    fig.subplots_adjust(hspace=0.5)
    plot_violins(data_frame.replace([np.inf, -np.inf], np.nan).dropna(), ax=axes[0])
    axes[0].set_ylim(-3, 5)
    axes[0].title.set_text("basal dendrites")

    if any(i.type == nm.NeuriteType.apical_dendrite for i in original_cells[0].neurites) and any(
        i.type == nm.NeuriteType.apical_dendrite for i in diametrized_cells[0].neurites
    ):
        data = get_features_all(
            original_cells,
            diametrized_cells,
            flist=VIOLIN_FEATURES_LIST,
            neurite_type=APICAL_DENDRITE,
        )

        data_frame = transform2DataFrame(data, pop_names, flist=VIOLIN_FEATURES_NAME)
        plot_violins(data_frame.replace([np.inf, -np.inf], np.nan).dropna(), ax=axes[1])
        axes[1].set_ylim(-3, 5)
        axes[1].title.set_text("apical dendrites")

    if with_axon:
        data = get_features_all(
            original_cells,
            diametrized_cells,
            flist=VIOLIN_FEATURES_LIST,
            neurite_type=AXON,
        )
        data_frame = transform2DataFrame(data, pop_names, flist=VIOLIN_FEATURES_NAME)
        plot_violins(data_frame.replace([np.inf, -np.inf], np.nan).dropna(), ax=axes[2])
        axes[2].set_ylim(-3, 5)

    return mtype, fig



"""Utils functions."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import logging
import os
from collections import defaultdict
from pathlib import Path
from xml.etree import ElementTree as ET

import numpy as np
import pandas as pd
from neurom import COLS

from diameter_synthesis.exception import DiameterSynthesisError

L = logging.getLogger(__name__)


def _create_morphologies_dict_dat(morph_path, mtypes_file="neurondb.dat"):
    """Create dict to load the morphologies from a directory, with dat file.

    Args:
        morph_path (str): path to morphologies.
        mtype_file (str): path to dat file.

    Returns:
        dict: dictionary of morphologies keyed by mtypes.
    """
    # pylint wrongly determines `morph_name` as TextFileReader
    # pylint: disable=no-member
    morph_name = pd.read_csv(mtypes_file, sep=r"\s+", header=None)
    name_dict = defaultdict(list)
    if not morph_name.empty:
        first_name = morph_name.loc[0, 0]  # pylint: disable=no-member
        file_list = Path(morph_path).glob(first_name + "*")
        try:
            ext = next(file_list).suffix
        except StopIteration as e:
            raise DiameterSynthesisError(f"Could not find a file starting with {first_name}") from e
    for morph in morph_name.values:  # pylint: disable=no-member
        name_dict[morph[2]] += [Path(morph_path) / (morph[0] + ext)]
    return name_dict


def _create_morphologies_dict_folder(morph_path):
    """Create dict to load the morphologies from a directory, from folders.

    Args:
        morph_path (str): path to morphologies.

    Returns:
        dict: dictionary of morphologies keyed by mtypes.
    """
    name_dict = defaultdict(list)
    for mtype in Path(morph_path).iterdir():
        for fname in mtype.iterdir():
            if fname.suffix in [".h5", ".asc", ".swc"]:
                name_dict[mtype.parts[-1]] += [fname]
    return name_dict


def _create_morphologies_dict_all(morph_path):
    """Create dict to load the morphologies from a directory, all together.

    Args:
        morph_path (str): path to morphologies.

    Returns:
        dict: dictionary of morphologies with single key.
    """
    name_dict = {"generic_type": []}
    for fname in Path(morph_path).iterdir():
        if fname.suffix in [".h5", ".asc", ".swc"] and fname.exists():
            name_dict["generic_type"] += [fname]

    return name_dict


def create_morphologies_dict(morph_path, mtypes_file=None):
    """Create dict to load the morphologies from a directory, by mtype.

     Args:
        morph_path (str): path to morphologies.
        mtype_file (str): path to dat file.

    Returns:
        dict: dictionary of morphologies with single key.
    """
    if mtypes_file is None and next(Path(morph_path).iterdir()).is_dir():
        L.info("found folder structure per mtype")
        return _create_morphologies_dict_folder(morph_path)
    if mtypes_file is not None:
        # mtype_path = Path(morph_path) / mtypes_file
        mtype_path = Path(mtypes_file)
        if mtype_path.exists():
            if mtype_path.suffix == ".dat":
                L.info("found dat file")
                return _create_morphologies_dict_dat(morph_path, mtypes_file=mtypes_file)
            raise DiameterSynthesisError(
                f"neurondb file format {mtype_path.suffix} not implemented"
            )

    L.info("use all files as single mtype")
    return _create_morphologies_dict_all(morph_path)


def _set_diameters(section, diameters):
    """Set diameters.

    Args:
        section (neurom.Section): section to diametrize.
        diameters (list/ndarray): diameters.
    """
    section.morphio_section.diameters = diameters


def _get_mean_diameter(section):
    """Section mean diameter.

    It is obtained by averaging the segment truncated cone diameters and weighting them by their
    length. (neurom only)

    Args:
        section (neurom.Section): section to consider.

    Returns:
        float: mean diameter of the section.
    """
    segment_lengths = np.linalg.norm(
        section.points[1:, COLS.XYZ] - section.points[:-1, COLS.XYZ], axis=1
    )
    segment_mean_diams = section.points[1:, COLS.R] + section.points[:-1, COLS.R]
    return np.sum(segment_mean_diams * segment_lengths) / segment_lengths.sum()


def get_all_diameters(neuron):
    """Get all neuron diameters (morphio only).

    Args:
        neuron (morphio.mut.Morphology): neuron to consider

    Returns:
        list: all diameters
    """
    return [np.array(section.diameters) for section in neuron.iter()]


def set_all_diameters(neuron, all_diameters):
    """Set all neuron diameters (morphio only).

    Args:
        neuron (morphio.mut.Morophology): neuron to consider.
        all_diameters (list): list of section diameters.
    """
    for diameters, section in zip(all_diameters, neuron.iter()):
        section.diameters = diameters


def _get_diameters(section):
    """Get diameters (neurom only).

    Args:
        section (neurom.Section): section to consider.

    Return:
        list: diameters of section
    """
    return section.points[:, COLS.R] * 2.0


def redefine_diameter_section(section, diam_ind, diam_new):
    """Replace given diameters at indices diam_ind with values diam_new (morphio only).

    Args:
        section (neurom.Section): section to consider.
        diam_ind (list): indices of diameters, or points on the section.
        diam_new (list): corresponding diameters.
    """
    diameters = section.diameters
    diameters[diam_ind] = diam_new
    section.diameters = diameters


#######################################################################
# old parser functions (will be removed or integrated at some points) #
#######################################################################
def _create_morphologies_dict_json(
    morph_path,
    mtypes_file="neuronDB.xml",
    prefix="",
):
    """Create dict to load the morphologies from a directory, with json."""
    with open(mtypes_file, "r", encoding="utf-8") as filename:
        morph_name = json.load(filename)

    name_dict = {}
    for fname in os.listdir(morph_path):
        filepath = Path(morph_path) / fname
        if fname.endswith((".h5", ".asc", ".swc")) and filepath.exists():
            mtype = morph_name[Path(fname).stem]
            if mtype in name_dict:
                name_dict[mtype] += [prefix + fname]
            else:
                name_dict[mtype] = [prefix + fname]

    return name_dict


def _create_morphologies_dict_xml(
    morph_path,
    mtypes_file="neuronDB.xml",
    ext=".asc",
    prefix="",
):
    """Create dict to load the morphologies from a directory, from xml."""
    filedb = ET.parse(morph_path + mtypes_file)
    root = filedb.findall("listing")[0]
    morphs = root.findall("morphology")

    name_dict = {}
    if len(morphs) > 0:
        for morph in morphs:
            try:
                # Define mtypes
                mtype = morph.find("mtype").text
                # Define subtypes (if they exist)
                if morph.find("msubtype").text:
                    mtype = mtype + ":" + morph.find("msubtype").text

                if mtype not in name_dict:
                    name_dict[mtype] = [prefix + morph.find("name").text + ext]
                elif mtype in name_dict:
                    name_dict[mtype] += [prefix + morph.find("name").text + ext]

            except Exception as exc:  # pylint: disable=broad-except
                L.exception("Failed to process %s", exc)

    return name_dict



"""Main functions to learn and generate diameters."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import logging
import multiprocessing
import os
from pathlib import Path

import neurom as nm
import numpy as np
from morphio.mut import Morphology
from tqdm import tqdm

from diameter_synthesis import utils
from diameter_synthesis.build_diameters import build as build_diameters
from diameter_synthesis.build_models import build as build_model
from diameter_synthesis.plotting import plot_distribution_fit

L = logging.getLogger(__name__)
logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))


class NumpyEncoder(json.JSONEncoder):
    """Class to encode numpy arrays."""

    def default(self, o):  # pylint: disable=method-hidden
        """Numpy encoder."""
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, np.floating):
            return float(o)
        if isinstance(o, np.integer):
            return int(o)
        return json.JSONEncoder.default(self, o)


def plot_models(morphologies, config, models_params, models_data, ext=".png"):
    """Plot the models.

    Args:
        morphologies (dict): a dict with mtype->[morphologies].
        config (dict): the config to use.
        models_params (dict): the models parameters.
        models_data (dict): the models data.
        ext (str): the file extension used to export the figures.
    """
    L.info("Plot the fits...")

    if not ext.startswith("."):
        ext = f".{ext}"

    if not Path(config["fig_folder"]).exists():
        os.mkdir(config["fig_folder"])

    for mtype in tqdm(morphologies):
        if not (Path(config["fig_folder"]) / mtype).exists():
            os.mkdir(Path(config["fig_folder"]) / mtype)
        for model in config["models"]:
            fit_tpes = models_data[model][mtype]
            model_data = models_data[model][mtype]
            model_param = models_params[model][mtype]

            for fit_tpe in fit_tpes:
                fig_name = Path(config["fig_folder"]) / mtype / fit_tpe
                plot_distribution_fit(
                    model_data[fit_tpe],
                    model_param[fit_tpe],
                    config["neurite_types"],
                    fig_name=fig_name,
                    ext=ext,
                )


def _build_all_models(morphologies, config, plot=False, ext=".png"):
    """Build all the models in the list of models."""
    models_params = {}
    models_data = {}
    for model in config["models"]:
        models_params[model] = {}
        models_data[model] = {}
        for mtype in tqdm(morphologies):
            config_model = config.copy()
            config_model["models"] = [model]
            models_params[model][mtype], models_data[model][mtype] = build_model(
                morphologies[mtype], config_model, with_data=True
            )

    if plot:
        plot_models(morphologies, config, models_params, models_data, ext=ext)

    return models_params


def run_models(config_file, plot, ext=".png"):
    """Run the model extraction from config file.

    Args:
        config_file (str): the path to the configuration file.
        plot (bool): plot the models once they are built.
        ext (str): the file extension used to export the plots.
    """
    with open(config_file, "r", encoding="utf-8") as filename:
        config = json.load(filename)

    L.info("Loading morphologies...")
    morphologies_dict = utils.create_morphologies_dict(
        config["morph_path"],
        mtypes_file=config.get("mtypes_file", None),
    )

    morphologies = {
        mtype: [nm.load_morphology(i) for i in morphologies_dict[mtype]]
        for mtype in tqdm(morphologies_dict)
    }

    L.info("Extracting model parameters...")
    models_params = _build_all_models(morphologies, config, plot=plot, ext=ext)

    with open(config["models_params_file"], "w", encoding="utf-8") as json_file:
        json.dump(models_params, json_file, sort_keys=True, indent=4, cls=NumpyEncoder)


class DiameterWorker:
    """Worker for building diameters.

    Args:
        model (str): The model to use.
        models_params (dict): The parameters of the models containing the model name as key.
        config (dict): The configuration to use.
    """

    def __init__(self, model, models_params, config):
        """Init function."""
        self.model = model
        self.models_params = models_params[model]
        self.config = config

    def __call__(self, neuron_input):
        """Call function."""
        fname = neuron_input[0]
        mtype = neuron_input[1]

        neuron = Morphology(fname)

        build_diameters(
            neuron,
            self.config["neurite_types"],
            self.models_params[mtype],
            self.config,
        )

        if not Path(self.config["new_morph_path"]).exists():
            os.mkdir(self.config["new_morph_path"])

        if fname.parts[-2] == mtype:
            if not (Path(self.config["new_morph_path"]) / fname.parts[-2]).exists():
                os.mkdir(Path(self.config["new_morph_path"]) / fname.parts[-2])
            neuron.write(Path(self.config["new_morph_path"]) / fname.parts[-2] / fname.name)
        neuron.write(Path(self.config["new_morph_path"]) / fname.name)


def run_diameters(config_file, models_params_file):
    """Build new diameters from config file and diameter model.

    Args:
        config_file (str): the path to the configuration file.
        models_params_file (str): the path to the file containing the model parameters.
    """
    with open(config_file, "r", encoding="utf-8") as filename:
        config = json.load(filename)

    with open(models_params_file, "r", encoding="utf-8") as filename:
        models_params = json.load(filename)

    for model in config["models"]:
        L.info("Generating diameter with model %s", model)

        morphologies_dict = utils.create_morphologies_dict(
            config["morph_path"],
            mtypes_file=config.get("mtypes_file", None),
        )

        worker = DiameterWorker(model, models_params, config)

        all_neurons = [
            [neuron, mtype] for mtype in morphologies_dict for neuron in morphologies_dict[mtype]
        ]

        with multiprocessing.Pool(config.get("n_cpu", 1)) as pool:
            list(tqdm(pool.imap(worker, all_neurons), total=len(all_neurons)))


def diametrize_single_neuron(neuron, config=None, apical_point_sec_ids=None):
    """Diametrize single neuron by learning diameter model from it.

    Args:
        neuron (mophio.mut.Morphology): neuron to consider.
        config (dict): dict with entry 'model' and 'diameters' with corresponding dicts, if None,
            default dict will be used.
        apical_point_sec_ids (list): list of apical points if any.
    """
    if config is None:
        config = {
            "model": {
                "taper": {"max": 1e-06, "min": -0.1},
                "terminal_threshold": 2.0,
                "models": ["generic"],
                "neurite_types": ["basal", "apical"],
            },
            "diameters": {
                "models": ["generic"],
                "n_samples": 1,
                "neurite_types": ["basal", "apical"],
                "seed": 0,
                "trunk_max_tries": 100,
            },
        }

    model_params = build_model([nm.load_neuron(neuron)], config["model"])
    if apical_point_sec_ids is not None:
        model_params["apical_point_sec_ids"] = apical_point_sec_ids
    build_diameters(neuron, ["basal", "apical"], model_params, config["diameters"])



"""Fetch some barrel morphologies from neuromorpho.org

Copyright (C) 2021  Blue Brain Project, EPFL

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import pandas as pd
import os
import re
import json
from urllib.request import urlopen, Request
from pathlib import Path

NEUROMORPHO_URL = "http://neuromorpho.org"


def get_swc_by_neuron_index(neuronIndex, folder="morphologies"):
    """Download a neuron by index and store it into a SWC file
    Keyword arguments:
    neronIndex -- the neuron index in the database

    Adapted from https://github.com/NeuroBox3D/neuromorpho/blob/master/rest_wrapper/rest_wrapper.py
    """

    url = "%s/api/neuron/id/%i" % (NEUROMORPHO_URL, neuronIndex)
    req = Request(url)
    response = urlopen(req)
    neuron_name = json.loads(response.read().decode("utf-8"))["neuron_name"]
    url = "%s/neuron_info.jsp?neuron_name=%s" % (NEUROMORPHO_URL, neuron_name)
    html = urlopen(url).read().decode("utf-8")
    p = re.compile(r"<a href=dableFiles/(.*)>Morphology File \(Standardized\)</a>", re.MULTILINE)
    m = re.findall(p, html)
    for match in m:
        file_name = match.replace("%20", " ").split("/")[-1]
        response = urlopen("%s/dableFiles/%s" % (NEUROMORPHO_URL, match))
        with open(folder + "/" + file_name, "w") as f:
            f.write(response.read().decode("utf-8"))


if __name__ == "__main__":
    brainRegion = "barrel"
    cell_type = "interneuron"
    numNeurons = 500

    url = "%s/api/neuron/select?q=brain_region:%s&size=%i" % (
        NEUROMORPHO_URL,
        brainRegion,
        numNeurons,
    )
    req = Request(url)
    response = urlopen(req)
    neurons = json.loads(response.read().decode("utf-8"))
    df = pd.DataFrame(neurons["_embedded"]["neuronResources"])
    df["type"] = df["cell_type"].apply(lambda t: t[-1])
    df = df[df["type"] == "principal cell"]
    folder = "morphologies"
    if not Path(folder).exists():
        os.mkdir(folder)
    for gid in df.index:
        get_swc_by_neuron_index(df.loc[gid, "neuron_id"], folder=folder)



"""Test the build_models module."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pytest

from diameter_synthesis import build_models
from diameter_synthesis.exception import DiameterSynthesisError

from .testing_tools import compare_dicts


def test_build(single_pop, model_params, model_data, empty_build_result):
    """Test the build function."""
    # Test with generic model
    config = {
        "models": ["generic"],
        "neurite_types": ["basal", "apical"],
        "terminal_threshold": 2,
        "taper": {"max": 1e-6, "min": -0.01},
    }
    res = build_models.build(single_pop, config, with_data=True)
    res_models_params = build_models.build(single_pop, config, with_data=False)

    assert len(res) == 2
    compare_dicts(res, [model_params, model_data], precision=3)
    compare_dicts(res_models_params, model_params, precision=3)

    # Test with astrocyte model
    config_astrocyte = {
        "models": ["astrocyte"],
        "neurite_types": ["basal", "apical"],
        "terminal_threshold": 2,
        "taper": {"max": 1e-6, "min": -0.01},
    }
    res_astrocyte = build_models.build(single_pop, config_astrocyte, with_data=True)

    assert len(res_astrocyte) == 2
    compare_dicts(res_astrocyte, [model_params, model_data], precision=3)

    # Test with empty population
    res_empty = build_models.build([], config_astrocyte, with_data=True)
    assert res_empty == empty_build_result

    # Test with unknown model (should raise a DiameterSynthesisError exception)
    bad_config = {
        "models": ["UNKNOWN"],
        "neurite_types": ["basal", "apical"],
        "terminal_threshold": 2,
        "taper": {"max": 1e-6, "min": -0.01},
    }
    with pytest.raises(DiameterSynthesisError):
        build_models.build(single_pop, bad_config, with_data=True)



"""Setup for the diameter-synthesis package.

Copyright (C) 2021  Blue Brain Project, EPFL

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from setuptools import find_packages
from setuptools import setup

# Read the contents of the README file
with open("README.md", encoding="utf-8") as f:
    README = f.read()

reqs = [
    "click>=7.0",
    "jsonschema>=3",
    "matplotlib>=2.2.0",
    "morphio>=2.3.4",
    "neurom>=3.0,<4.0",
    "numpy>=1.15.0",
    "pandas>=0.24.0",
    "scipy>=0.13.3",
]

doc_reqs = [
    "m2r2",
    "sphinx",
    "sphinx-bluebrain-theme",
    "sphinx-jsonschema",
    "sphinx_click",
]

test_reqs = [
    "diff_pdf_visually>=1.5.1",
    "decorator",
    "matplotlib",
    "mock",
    "pytest",
    "pytest-cov",
    "pytest-html",
    "pytest-xdist",
]

setup(
    name="diameter-synthesis",
    author="Blue Brain Project, EPFL",
    description="Diametrize cells",
    long_description=README,
    long_description_content_type="text/markdown",
    license="GPLv3",
    url="https://github.com/BlueBrain/diameter-synthesis",
    project_urls={
        "Tracker": "https://github.com/BlueBrain/diameter-synthesis/issues",
        "Source": "https://github.com/BlueBrain/diameter-synthesis",
    },
    install_requires=reqs,
    extras_require={
        "docs": doc_reqs,
        "plot": "seaborn>=0.11.1",
        "test": test_reqs,
    },
    tests_require=test_reqs,
    entry_points={
        "console_scripts": ["diameter-synthesis=diameter_synthesis.cli:cli"],
    },
    python_requires=">=3.8",
    use_scm_version=True,
    setup_requires=[
        "setuptools_scm",
    ],
    packages=find_packages(exclude="tests"),
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)
